<?php 
namespace Concrete\Package\GntDotRender;
defined('C5_EXECUTE') or die("Access Denied.");
/**
 *-----------------------------------------------------------------------------*
 * License: Concrete5 marketplace :                                            *
 * http://www.concrete5.org/help/legal/commercial_add-on_license/              *
 *                                                                             *
 * @author (Florian Delizy)                                                    *
 *                                                                             *
 * Copyright(C) 2014 Florian Delizy <florian.delizy@gmail.com>                 *
 *-----------------------------------------------------------------------------*
 */

use Events;
use Package;
use BlockType;
use SinglePage;
use Page;
use Database;
use Log;
use Config;

use Concrete\Core\Routing\Router;
use Concrete\Core\Support\Facade\Route;

use \Concrete\Core\Page\Controller\DashboardPageController;

class Controller extends Package 
{
	
	protected $pkgHandle = 'gnt_dot_render';
	protected $appVersionRequired = "5.7.0.4";
	protected $pkgVersion = "2.0.3";

	public function getPackageDescription() { return t( "Render a dot graph" ); }
	public function getPackageName() { return t( "Dot Render" ); }

	public function install()
	{
		$pkg = parent::install();

		BlockType::installBlockTypeFromPackage( "gnt_dot_render", $pkg );

        $co = $this->getConfig();

		$co->save("dot.path", '/usr/bin/dot' );
		$co->save("dot.options", '' );
		$co->save("dot.template", "%GRAPHMODE% %GRAPHID%\n{\n\tbgcolor=\"transparent\";\nnode[style=\"filled\",fillcolor=\"white\"];\n\t%BLOCKCODE%\n}\n" );

		$co->save('cache.enable', 1 );

		$co->save('cache.path', \Config::get( 'concrete.cache.directory') );

		// Dotrender properties :
		$dash = SinglePage::add('/dashboard/system/environment/dotrender', $pkg);
		$dash->update(array('cName' => t('Dot Render Settings'), 'cDescription' => t('Dot Render configuration (paths, options, template)')));

		// Dotrender cache :
		$dash = SinglePage::add('/dashboard/system/optimization/dotrender', $pkg);
		$dash->update(array('cName' => t('Dot Render Cache'), 'cDescription' => t('Dot Render cache control')));
	}

	// Configuration accessors

	public function getCachePath() { return $this->getConfig()->get( "cache.path" ); }
	public function getUseCache()  { return $this->getConfig()->get( "cache.enable" ) != 0; }
	public function getDotPath() { return $this->getConfig()->get( "dot.path" ); }
	public function getDotOpts() { return $this->getConfig()->get( "dot.options" ); }
	public function getDotTemplate() { return $this->getConfig()->get( "dot.template" ); }

	// Configuration Setters, checks that the path is a writable dir
	public function setCachePath( $path )
	{
		if ( !is_dir($path) ) throw new Exception( t( "%s is not a valid directory", $path ) );
		if ( !is_writable( $path ) ) throw new Exception( t( "%s is not writable", $path ) );
		$this->getConfig()->save( "cache.path", $path );
	}

	public function setUseCache( $use = true )  { $this->getConfig()->save( "cache.enable", $use ? 1 : 0 ); }

	// set the dot cmd path, checks that the command is a valid executable
	public function setDotPath( $path )
	{
		if ( ! file_exists( $path ) )   throw new Exception( t('%s is not a valid executable (file does not exist)', $path ) );
		if ( ! is_executable( $path ) ) throw new Exception( t( "%s is not a valid executable (can not execute)", $path ) );
		$this->getConfig()->save( "dot.path", $path );
	}
	public function setDotOpts( $opts ) { $this->getConfig()->save( "dot.options", $opts ); }
	public function setDotTemplate( $template ) { $this->getConfig()->save( "dot.template", $template ); }

	public function on_start()
	{
		Events::addListener( 'on_cache_flush', array( $this, 'flushCache' ) );

		$this->registerRoutes();
	}

	public function registerRoutes()
	{

		Route::register( Router::route( array( '/ajax/dot_image',     'gnt_dot_render') ), '\Concrete\Package\GntDotRender\Controller\Ajax\DotImage::view',    "GntDotRenderImage" );
		Route::register( Router::route( array( '/ajax/dot_image_tag', 'gnt_dot_render') ), '\Concrete\Package\GntDotRender\Controller\Ajax\DotImageTag::view', "GntDotRenderImageTag" );
	}

	public function flushCache( $event )
	{
        $pattern = $this->getCachePath() . "/gnt_dotrender_cache_*";
		$list = glob( $pattern);
		if (!is_array($list) || count( $list ) == 0 ) return;
        foreach ( $list as $file ) @unlink($file);
	}

	public function uninstall() 
	{
		$this->flushCache(null);

		parent::uninstall();

		$db = Database::getActiveConnection();
		$db->Execute( 'DROP TABLE IF EXISTS btDotRender' );
	}
}

// vim: set noexpandtab ts=4 :
