<?php  
namespace Concrete\Package\GntDotRender\Src;
defined('C5_EXECUTE') or die("Access Denied.");

/**
 *-----------------------------------------------------------------------------*
 * License: Concrete5 marketplace :                                            *
 * http://www.concrete5.org/help/legal/commercial_add-on_license/              *
 *                                                                             *
 * @author (Florian Delizy)                                                    *
 *                                                                             *
 * Copyright(C) 2014 Florian Delizy <florian.delizy@gmail.com>                 *
 *-----------------------------------------------------------------------------*
 */

use Log;
use Package;
use Permissions;
use URL;
use Block;
use Core;

class DotRenderer
{
    const TYPE_PNG     = 0;
    const TYPE_JPG     = 1;
    const TYPE_RAW     = 2;
    const TYPE_CMAPX   = 3;
    const TYPE_GIF     = 4;
    const TYPE_MAX     = 5; // if > 10 must change getRestApiDesc

    const MODE_DIRECTED   = 0;
    const MODE_UNDIRECTED = 1;
    const MODE_MAX        = 2; // if > 10 must change getRestApiDesc

    private $_type;
    private $_code;
    private $_mode;

    private $_tool;
    private $_opts;

    private $_lastError = '';

    private $_uniquify = false;
    private $_useCache = false;

    private $_passDotPath = false;
    private $_passDotOpts = false;
    private $_passDotTemplate = false;

    private $_templateMode = false;
    private $_renderCmap = false;
    private $_blockID = null;

    private $_hash = null;
    private $_baseCachePath = '';

    protected static $_mimeTypes = array (
        self::TYPE_PNG     => "image/png",
        self::TYPE_JPG     => "image/jpg",
        self::TYPE_RAW     => "text/plain",
        self::TYPE_CMAPX   => "text/html",
        self::TYPE_GIF     => "image/gif",
    );

    protected static $_extensions = array (
        self::TYPE_PNG     => "png",
        self::TYPE_JPG     => "jpg",
        self::TYPE_RAW     => "dot",
        self::TYPE_CMAPX   => "cmap",
        self::TYPE_GIF     => "gif",
    );

    protected static $_modes = array (
        self::MODE_DIRECTED   => 'digraph',
        self::MODE_UNDIRECTED => 'graph',
    );

    public function getMime( $type = NULL ) { return self::$_mimeTypes[is_null($type)?$this->getType():$type]; }
    public function getExtension( $type = NULL ) { return self::$_extensions[is_null($type)?$this->getType():$type]; }
    public function getGraphTag( $mode = NULL ) { return self::$_modes[is_null($mode)?$this->getMode():$mode]; }

    protected function _getHash()
    {
        // I use a md5 hash here, I know that there is a _possibity_ of collision, but it is somewhat
        // low enough (< 1/2^100) to take the risk and it is 25% faster on average than a sha1 hash for instance.
        if ( is_null( $this->_hash ) ) $this->_hash = md5($this->getCode());
        return $this->_hash;
    }

    public function clearRelatedCache()
    {
        // For the PRB:
        //
        // Clear all cache belonging to that dot object .... actually ... not only :
        // Since the dot code used to build the image determine the hash used in the cache filename, 
        // in case two blocks share the same dot code (although unlikely), then 
        // modifying one of theses blocks will then clear the cache of both blocks 
        // (since actually, there is only one cache for both)
        //
        // The other alternative was to use a different hash for blocks, but I like the idea to
        // hash the code so that in return, two code sharing the same code will also share the same cache (and then
        // the image will only get built once).
        //
        // In other words, this is not a bug, this is a feature, /or close enough/ (C) :)

        $hash = $this->_getHash();
        $pattern = $this->_baseCachePath . "/gnt_dotrender_cache_${hash}*";
        $list = glob( $pattern );

        if (!is_array($list) or count( $list ) == 0 ) return;
        foreach ( $list as $file ) @unlink($file);
    }

    public function getCacheFileName( $type = null, $cmapId = '' )
    {
        if ( is_null($type) ) $type = $this->getType();
        $hash = $this->_getHash($type);
        $ext = $this->getExtension($type);
        return $this->_baseCachePath . "/gnt_dotrender_cache_$hash$cmapId.$ext";
    }

    public function getCacheFileNameIfExists( $type = null, $cmapId = '' )
    {
        if ( is_null($type) ) $type = $this->getType();
        $filename = $this->getCacheFileName($type, $cmapId);
        if ( file_exists( $filename ) ) return $filename;
        return false;
    }

    public function useCache() { return $this->_useCache; }

    public function __construct( $type = self::TYPE_PNG, $mode = self::MODE_DIRECTED )
    {
        $pkg = Package::getByHandle('gnt_dot_render');
        $this->setType( $type );
        $this->setToolPath( $pkg->getDotPath(), false );
        $this->setToolOpts( $pkg->getDotOpts(), false );
        $this->setTemplate( $pkg->getDotTemplate(), false );
        $this->_baseCachePath = $pkg->getCachePath();
        $this->_useCache = $pkg->getUseCache();
        $this->_mode = $mode;
    }

    // Accessors :
    public function setToolPath( $path, $pass = true )    
    { 
        if ( $path == $this->_tool ) return;
        $this->_tool = $path;
        $this->_passDotPath = $pass;
    }


    public function setTemplate( $template, $pass = true) 
    { 
        if ( $this->_template == $template ) return;
        $this->_template = $template;
        $this->_passDotTemplate = $pass;
    }

    public function setToolOpts( $opts, $pass = true )     
    {
        if ( $this->_opts == $opts ) return;
        $this->_opts = $opts;
        $this->_passDotOpts = $pass;
    }

    public function getTemplateCode( $mode = null )
    {
        $mode = is_null( $mode ) ? $this->getMode() : $mode;
        $link = $mode == self::MODE_DIRECTED ? '->' : '--';

        $code =' 
        rankdir="LR";
        size="3.6,1.56";
        n [shape=diamond,label="",fillcolor="#5fbcd3",style="filled"];
        D [URL="http://www.graphviz.org",fillcolor="#5fbcd3",style="filled"];
        o [URL="http://www.graphviz.org",style="filled",fillcolor="#5fbcd3"];
        t [URL="http://www.graphviz.org",style="filled",fillcolor="#5fbcd3"];
        D '.$link.' o;
        o '.$link.' n [constraint=false];
        D '.$link.' n;
        n '.$link.' t;
        ';
        return $code;
    }
    public function setTemplateMode()
    {
        $this->_templateMode = true;
    }
    public function setMode( $mode ) { $this->_mode = $mode; }
    public function getMode() { return $this->_mode; }
    public function setCode( $code )
    { 
        $pattern = '/^\s*(strict\s*)?(di)?graph\s([\d\w])?\s*\{/';
        if ( preg_match( $pattern, $code ) )
        {
            $code = preg_replace( array( $pattern, '/}\s*$/' ), array( '', '' ), $code );
        }
        $this->_code = $code;
    }
    public function setType( $type ) { $this->_type = $type; }
    public function getType() { return $this->_type; }
    public function renderCmap ( $render ) { $this->_renderCmap = $render; }
    public function uniquify()
    { 
        $this->_uniquify = true;
        $this->_useCache = false;
    }

    /**
     * @fn getAvailableTypes ()
     * get an array of supported types with a description
     * @return array( 'type' => 'name' )
     */
    public static function getAvailableTypes ()
    {
        return array (
            self::TYPE_PNG => t( 'PNG Image' ),
            self::TYPE_JPG => t( 'Jpeg Image' ),
            self::TYPE_GIF => t( 'Gif Image' ),
            self::TYPE_RAW => t( 'RAW dot program' ),
        );
    }

    public static function getAvailableModes ()
    {
        return array (
            self::MODE_DIRECTED   => 'Directed Graph (digraph)',
            self::MODE_UNDIRECTED => 'Undirected Graph (graph)'
        );
    }

    /**
     * @fn image( $bID = null, $cmapId = null )
     * output the image tag corresponding to the graph method
     * @param $bID (boolean) the block ID
     * @param $renderCmap (boolean) if true, render a cmap too
     */
    public function image( $bID = null, $renderCmap = null )
    {
        if ( $this->_type == self::TYPE_RAW )
        {
            echo '<pre class="dotrender raw">';
            echo $this->getCode();
            echo '</pre>';
            return;
        }

        if ( $this->_type == self::TYPE_CMAPX )
        {
            // This is voluntary forbidden, use render instead if you know what you are doing
            echo '<div class="ccm-error">"' . t("Can not render cmapx without rendering image") . '</div>';
            return;
        }

        if ( is_null ($renderCmap ) ) $renderCmap = $this->_renderCmap;

        $imageTarget = URL::route( array( '/ajax/dot_image', 'gnt_dot_render' ) );

        if ( is_null( $bID ) ) $bID = $this->_blockID;

        if ( $renderCmap )
        {
            $cmapId = is_null($bID) ? "map_" . uniqid( $bID ) : "gnt_dotrender_blk_$bID";
            $this->render( self::TYPE_CMAPX, $cmapId );
            $map = "usemap=\"$cmapId\"";
        }

        $param = "type=" . $this->_type;
        if ( !is_null( $bID ) ) $param .= "&bID=$bID";
        else if ( $this->_templateMode ) $param .= "&template=1";
        else
        {
            // TODO: Compress the code to be sent for rendering ?
            $code=urlencode(base64_encode( $this->_code ));
            $param .= "&code=$code";
        }

        $param .= "&mode=" . $this->getMode();

        if ( $this->_passDotPath ) $param .= "&dotpath=" . urlencode( $this->_tool );
        if ( $this->_passDotOpts ) $param .= "&dotopts=" . urlencode( $this->_opts );
        if ( $this->_passDotTemplate ) $param .= "&dottemplate64=" . urlencode( base64_encode( $this->_template ) );
        if ( $this->_uniquify ) $param .= "&uniquify=" . urlencode ( uniqid() );
        
        $url =  htmlentities( "$imageTarget?$param" );

        echo "<img $map src=\"$url\" alt=\"dot graph generated image\"/>";
    }

    public function getCode( $id = null )
    {
        $code = $this->_templateMode ? $this->getTemplateCode() : $this->_code;

        $graphID = is_null($id) ? 'G' : "\"$id\"";

        $patterns = array( '/%BLOCKCODE%/', '/%GRAPHID%/', '/%GRAPHMODE%/' );
        $ret = preg_replace( $patterns,  array( $code, $graphID, $this->getGraphTag() ), $this->_template );
        return $ret;
    }

    public function render( $outType = null, $cmapId = null )
    {
        // passthru call, make it as fast as possible :
        if ( !$this->useCache() ) return $this->_render( $outType, $cmapId );

        // caching enabled
        $txtId = is_null( $cmapId ) ? '' : $cmapId;
        $filename = $this->getCacheFileNameIfExists();
        if ( !$filename )
        {
            ob_start();
            $r = $this->_render( $outType, $cmapId );
            $content = ob_get_clean();

            $filename = $this->getCacheFileName($outType, $txtId );
            $f = fopen( $filename, 'w' );
            if ( !is_resource($f) )
            {
                Log::addEntry( t("Can not write cache file %s, disabling cache for this element", $filename), t('error') );
                $this->_useCache = false;
                return $this->_render( $outType, $cmapId );
            }

            fwrite( $f, $content );
            fclose( $f );
            echo $content;
            return $f;
        }

        // NOTE (mainly for the PRB):
        // Here I prefer using the direct call file_get_contents
        // instead of FileHelper::getContents() since the file is *always* local
        // and *MUST* be local. Since outputting errors at this point could Just
        // break the render and be very difficult to trace, just disable it
        echo @file_get_contents($filename);
        return true;
    }

    protected function _render( $outType = null, $cmapId = null )
    {
        $type = $outType;
        if ( is_null( $outType ) ) $type = $this->_type;

        if ( self::TYPE_RAW == $type )
        {
            echo $this->getCode();
            return true;
        }

        $out = array(
            self::TYPE_PNG   => 'png',
            self::TYPE_RAW   => '',
            self::TYPE_JPG   => 'jpg',
            self::TYPE_GIF   => 'gif',
            self::TYPE_CMAPX => 'cmapx',
        );

        $ttype = $out[$type];

        if ( $ttype ) $ttype = "-T$ttype";

        $cmd = $this->_tool . " $ttype $this->_opts";
        $opts = array ( 
                    0 => array( 'pipe', 'r' ),
                    1 => array( 'pipe', 'w' ),
                    2 => array( 'pipe', 'w' ),
                );

        $fh = Core::make( 'helper/file' );
        $process = proc_open( $cmd, $opts, $pipes, $fh->getTemporaryDirectory(), array() );

        if ( ! is_resource( $process ) ) return false;

        fwrite( $pipes[0], $this->getCode( $cmapId ) );
        fflush( $pipes[0] );
        fclose( $pipes[0] );

        $image = stream_get_contents( $pipes[1] );
        $this->_lastError = stream_get_contents($pipes[2] );

        fclose( $pipes[1] );
        fclose( $pipes[2] );
        $r = proc_close( $process );

        if ( 0 != $r )
        {
            if ( $this->_lastError ) throw new \Exception($this->_lastError);
            throw \Exception( t("Undocumented dot error :/") );
        }

        echo $image;

        return true;
    }

    public function getError() { return $this->_lastError; }


    public function setBlockId( $bID )
    {
        $this->_blockID = $bID;
        // retrieve the code from the block
        $block = Block::getByID( $bID );
        if ( !$block ) return false;
        $controller = $block->getController();
        $this->setCode( $controller->blockCode );
        $this->renderCmap( $controller->renderCmapX != 0 );
        $this->setMode( $controller->mode );
        return true;
    }


    public static function getRestApiDesc()
    {
        // requiredGroups are exclusive (one in each group)
        return array(
            'type'          => array( 'regex' => '[0-' . ( self::TYPE_MAX - 1 ) . ']+' ),
            'uniquify'      => array( 'regex' => '[01]?' ),
            'code'          => array( 'regex' => '.*', 'requiredGroup' => t('code definition') ),
            'plaincode'     => array( 'regex' => '.*', 'requiredGroup' => t('code definition') ),
            'renderCmap'    => array( 'regex' => '[01]?' ),
            'mode'          => array( 'regex' => '[0-' . ( self::MODE_MAX - 1 ) . ']' ),
            'dotpath'       => array( 'regex' => '.*' ), 
            'dotopts'       => array( 'regex' => '.*' ), 
            'dottemplate'   => array( 'regex' => '.*' ), 
            'dottemplate64' => array( 'regex' => '[0-9a-zA-Z]+' ),
            'bID'           => array( 'regex' => '[0-9]+', 'requiredGroup' => t('code definition') ),
            'template'      => array( 'regex' => '[01]?',  'requiredGroup' => t('code definition') ),
        );
    }

    /**
     * @fn optionSetIsValid( array $opts )
     * Check if the option set is valid, return an error message if not 
     *
     * @param opts (associative array) 
     * @return (mixed) true if the option set is valid, an error string if not
     */
    public static function optionSetIsValid( array $opts )
    {
        $proto          = self::getRestApiDesc();
        $requiredGroups = array();

        foreach( $proto as $option => $desc )
            if ( array_key_exists( "requiredGroup", $desc ) ) $requiredGroups[$desc['requiredGroup']] = false;

        foreach ($proto as $option => $desc )
        {
            if ( !array_key_exists( $option, $opts ) ) continue;
            $value = $opts[$option];

            if ( !preg_match( '/' . $desc['regex'] . '/', $value ) ) return t('%s option invalid format (%s)', $option, $value);

            if ( array_key_exists( 'requiredGroup', $desc ) ) 
            {
                $group = $desc['requiredGroup'];
                if ( $requiredGroups[$group] ) return t('Conflicting options');
                $requiredGroups[$desc['requiredGroup']] = true;
            }
        }

        $key = array_search( false, $requiredGroups);

        if ( false !== $key ) return t('Missing required option (%s)', $key );
        return true;
    }

    public function canSetOption( $option )
    {
        $restricted = array (
            'dotpath',
            'dotopts', 
            'dottemplate',
            'dottemplate64',
        );

        $page = Page::getByPath( 'dashboard/system/environment/dotrender' );
        $p = new Permissions( $page );

        // Since allowing caller to change the dotpath/options or template can be a heavy risk
        // We only allow user who could have modified the dashboard page in the first place
        // --> This is used to test parameters in the dashboard page

        if ( in_array( $option, $restricted) ) return $p->canWrite();

        return true;
    }

    public function setRestOption( $option, $value )
    {
        switch ( $option )
        {
            case 'type' :
            {
                $type = intval($value);
                if ( $type < 0 ) throw new \Exception ( t( 'Invalid type' ) );
                if ( $type > self::TYPE_MAX ) throw new \Exception( t( 'Unsupported type' ) );
                $this->setType($type);
                break;
            }

            case 'uniquify' :
                if ( '0' !== $value ) $this->uniquify();
                break;

            case 'bID' : 
                if ( !$this->setBlockId( $value ) ) throw new \Exception( t('Block not found') );
                break;

            case 'code':
                $this->setCode( base64_decode( $value ) );
                break;

            case 'plaincode':
                $this->setCode( $value );
                break;

            case 'renderCmap':
                $this->renderCmap( $value !== '0' );
                break;

            case 'mode' :
                $this->setMode( $value );
                break;

            case $req['dotpath'] :
                $this->setToolPath ( $value );
                break;

            case $req['dotopts'] :
                $this->setToolOpts ( $value );
                break;

            case 'dottemplate' :
                $this->setTemplate( $value );
                break;

            case 'dottemplate64':
                $this->setTemplate( base64_decode( $value ) );
                break;

            case 'template':
                if ($value !== '0') $this->setTemplateMode();
                break;

        }
    }

    public function setRestOptions( array $options )
    {
        foreach ( array_keys(self::getRestApiDesc()) as $option )
        {
            if ( array_key_exists( $option, $options ) ) $this->setRestOption($option, $options[$option]);
        }
    }

    /**
     * @fn getFromRequest( $req = $_REQUEST )
     * Create an appropriate object based on the request, throw Exeption( message ) on error
     * @param $req (array) the array containing the request (default to $_REQUEST)
     * @return (mixed) DotRenderer in case of success, string in case of error
     */
    public static function getFromRequest( array $req )
    {
        $r = self::optionSetIsValid($req);
        if ( true !== $r ) throw new \Exception($r);

        $dot = new static();
        $dot->setRestOptions( $req );
        return $dot;
    }


    public static function outputErrorImage( $message )
    {
        header( "Content-Type: " . self::$_mimeTypes[self::TYPE_PNG] );

        $ttf  = DIR_PACKAGES . "/gnt_dot_render/media/fonts/arial.ttf";
        $left = DIR_PACKAGES . "/gnt_dot_render/media/images/error_left.png";
        $right= DIR_PACKAGES . "/gnt_dot_render/media/images/error_right.png";
        $fontsz = 12;
        $titlesz = 20;

        $title = t("%s image error", t("Dot Render"));
        $rect = imagettfbbox( $fontsz, 0, $ttf, $message );
        $trect = imagettfbbox( $titlesz, 0, $ttf,  $title);

        if ( false === $rect ) 
        {
            Log::warning( t("Can not evaluate message image size:, message was: '%s'", $message ) );
            return;
        }

        $w = $rect[2] - $rect[0];
        $h = $rect[1] - $rect[7];

        $tw = $trect[2] - $trect[0];
        $th = $trect[1] - $trect[7];

        $leftRect  = getimagesize($left);
        $rightRect = getimagesize($right);

        $leftImg  = imagecreatefrompng( $left );
        $rightImg = imagecreatefrompng( $right );


        $imgw = max($tw,$w) + $leftRect[0] + $rightRect[0];
        $imgh = max($h+$th, $leftRect[1], $rightRect[1]);

        $img = imagecreatetruecolor( $imgw, $imgh );

        if (!$img)
        {
            Log::warning( t("Can not create error image, message was: '%s'", $message ) );
            return;
        }

        imagealphablending($img, false);
        $trans = imagecolorallocatealpha($img, 0, 0, 0, 0 );
        imagefill( $img, 0, 0, $trans );

        $bg = imagecolorat( $leftImg, $leftRect[0]-1, $leftRect[1]/2);

        $r = ($bg>>16) & 0xff;
        $g = ($bg>>8)  & 0xff;
        $b = ($bg>>0)  & 0xff;

        $col = array( $r, $g, $b );
        $maj = max($r, $g, $b);
        $majIdx = array_search( $maj, $col );

        $s = 128; // above is considered light 

        $fgCol = array( $r>$s?255:0, $g>$s?255:0,$b>$s?255:0);
        $fgCol[$majIdx] = $maj>$s ? $maj/2: 255-(255-$maj)/2;

        $fg = imagecolorallocate( $img, $fgCol[0], $fgCol[1], $fgCol[2] );

        imagefilledrectangle( $img, ($imgw-$w)/2, 0, ($imgw-$w)/2+$w, $imgh, $bg );
        imagerectangle( $img, $leftRect[0]+1, 0, $leftRect[0]+max($w,$tw)-2, $th+1, $fg );

        imagecopy( $img, $leftImg, 0, 0, 0, 0, $leftRect[0], $leftRect[1]);
        imagecopy( $img, $rightImg, $imgw-$rightRect[0], 0, 0, 0, $rightRect[0], $rightRect[1]);

        imagettftext( $img, $titlesz, 0, ($imgw-$tw)/2, $th-$trect[1] + 2, $fg, $ttf, $title);

        imagettftext( $img, $fontsz, 0, ($imgw-$w)/2, ($th + $imgh + $h - $rect[1])/2, $fg, $ttf, $message );

        imagepng( $img );
    }
}

